import serial
import os
from datetime import datetime
from kb_controller.helpers import kb_config, kb_central, microbit, event_poller

current_track_id = 0
department_id = 0
tracks_per_department = {}

def start():
    clear_console()
    print()
    print(f'  CONTROLLER STARTER OP {datetime.now()}')
    print('===================================================')
    print()

    # TODO: prompt for api_user/password on startup, and set this in kb_config
    # (and remove from the .ini file!!)

    microbit.initialize_connection()

    global current_track_id
    global department_id
    global tracks_per_department

    department_id = kb_central.get_department_id_from_token()
    tracks_per_department = kb_central.get_tracks()

    show_console_info()

    while True:

        if current_track_id > 0 and event_poller.thread is None:
            event_poller.start_thread(current_track_id)

        try:
            byteText = microbit.read_line()
            response = ''
            text = str(byteText, 'utf-8').strip()

            if not text == '':
                print('m:b <<< ' + text)

                if (text == 'quit'):
                    microbit.send_message('OK:QUIT')
                    break

                if (text.startswith('init:')):
                    track = text[5:]
                    if not track in tracks_per_department[department_id]['tracks']:
                        print(f'ERROR! Bane ID {track} findes ikke for afdeling {department_id}')
                        response = 'ERROR:INIT:INVALID-TRACK:' + track
                    else:
                        current_track_id = int(track)
                        response = 'OK:INIT:' + track
                        print_tracks()

                elif (current_track_id == 0):
                    print('Der er ikke tilknyttet en kuglebane til controlleren')
                    print('Brug kommandoen: "init:<<kuglebane-ID>>" for at tilknytte kuglebanen')
                    print("Dette kan du f.eks. programmere micro:bit'en til at gøre :-)")
                    print()
                    response = 'ERROR:NOT-INITIALIZED'

                elif (text == 'events'):
                    print()
                    print(f'Henter events for kuglebanen')
                    print('-----------------------------')
                    count = kb_central.get_number_of_events_for_track(current_track_id)
                    response = f'OK:EVENTS-TRACK:{current_track_id}:{count}'

                elif (text.startswith('send:')):
                    to_track = text[5:]
                    to_track_exists = False
                    for dep_id in tracks_per_department:
                        to_track_exists = to_track in tracks_per_department[dep_id]['tracks']
                        if to_track_exists:
                            break

                    if not to_track_exists:
                        print(f'ERROR! Bane ID {to_track} findes ikke!')
                        response = 'ERROR:START:INVALID-TRACK:' + to_track
                    else:
                        print()
                        print(
                            f'Sender START besked til "bane {to_track}"')
                        print('-------------------------------------')
                        kb_central.send_start_event(current_track_id, to_track)
                        response = 'OK:SEND:' + to_track

                elif (text.startswith('pop')):
                        print()
                        print(
                            f'"Popper" den ældste vent fra event-køen')
                        print('-------------------------------------')
                        events = kb_central.get_events_for_track(current_track_id)
                        if events is None or not 'events' in events:
                            response = 'ERROR:POP:INGEN-EVENTS'
                        else:
                            response = pop_event(events)

                else:
                    print('Ukendt kommando modtaget: ' + text)
                    response = 'ERROR:unknown-command'

            if not response=='':
                microbit.send_message(response)

        except serial.serialutil.SerialException as e:
            print()
            print("Der skete en fejl med den serielle forbindelse!")
            print(type(e))
            print(e)
            print(e.args)
            break

        except Exception as e:
            print()
            print("Der skete en fejl!")
            print(type(e))
            print(e)
            print(e.args)
            microbit.send_message('ERROR:Exception')

    microbit.close_connection()

    print()
    print(f'FARVEL! Controlleren lukker ned nu! {datetime.now()}')
    print()
    print('Venter på at tråden der poller kuglebane-centralen afslutter.')
    print(f'Der kan gå op til {kb_config.poll_interval_in_seconds} sekunder...')
    event_poller.signal_stop_thread()
    
def pop_event(events):
    # first event is always the oldest
    event_id_to_pop = list(events['events'])[0]
    event = events['events'][event_id_to_pop]

    command_id = event['command_id']
    from_department_id = event['from_department_id']
    from_track_id = event['from_track_id']
    from_track_name = event['from_track_name']

    print()
    print(f'Event ID: {event_id_to_pop}')
    print(f'  - kommando: {command_id}')
    print(f'  - modtaget fra afdeling: {from_department_id}')
    print(f'                kuglebane: {from_track_id} ("{from_track_name}")')
    print()

    kb_central.delete_event(event_id_to_pop)
    return f'OK:POP:{from_department_id}:{from_track_id}:{command_id}'

def clear_console():
    global department_id
    global tracks_per_department
    os.system('cls' if os.name == 'nt' else 'clear')

def show_console_info():
    print()
    print('============================================================')
    print('   M I C R O : B I T   S E R I A L   C O N T R O L L E R')
    print('============================================================')
    print(f'Kuglebane central........: {kb_config.kbc_host}')
    print(f'Controller for afdeling..: {department_id}')
    print('-----------------------------------------------------------')
    print()
    print('Kommandoer:')
    print()
    print('  init:<<kuglebane_id>>')
    print('     - fortæller hvilken kuglebane controlleren håndterer')
    print()
    print('  events')
    print('     - henter alle events for den kuglebane controlleren håndterer')
    print('       BEMÆRK: controlleren spørger automatisk efter disse events i')
    print('               det interval der er angivet i "polling_interval" i ')
    print('               kbc-config.ini filen')
    print()
    print('  pop')
    print('     - henter den ældste besked til den kuglebane controlleren håndterer')
    print("     - sender kommandoen fra beskeden til micro:bit'en")
    print("     - sletter beskeden fra kuglebane centralen")
    print()
    print('  send:<<to_track_id>>')
    print('     - sender en besked til en anden kuglebane om at starte kuglen')
    print()
    print('  quit')
    print('     - Afbryder controlleren')
    print()
    print('-----------------------------------------------------------')
    print()

def print_tracks():
    global current_track_id
    global department_id
    global tracks_per_department

    print()
    print('-----------------------------------------------------------')
    print('Tilgængelige kuglebaner:')
    print()
    print(f'  {department_id} (egen afdeling):')
    print('  -------------------------------------')
    own_tracks = tracks_per_department[department_id]['tracks']
    for track_id in own_tracks:
        is_controlled_track = (int(track_id) == current_track_id)
        print(f"    {track_id} - {own_tracks[track_id]}    {'[CONTROLLER]' if is_controlled_track else ''}")
    print('  -------------------------------------')

    for dep_id in tracks_per_department:
        if not dep_id == department_id:
            print()
            print(f'  {dep_id}:')
            tracks = tracks_per_department[dep_id]['tracks']
            for track_id in tracks:
                print(f"    {track_id} - {tracks[track_id]}")
    print('-----------------------------------------------------------')
    print()


if __name__ == '__main__':
    start()


import os
# from pathlib import Path
import configparser

script_path = os.path.dirname(os.path.realpath(__file__))
config = configparser.ConfigParser()

def save():
    with open('kbc-config.ini', 'w') as configfile:
        print("Gemmer konfigurationsfilen...")
        config.write(configfile)


# Intialize configuration values

try:
    with open('kbc-config.ini') as cfgFile:
        config.read_file(cfgFile)
except IOError:
    print()
    print('----------------------------------------------------------------')
    print("Ingen konfigurationsfil fundet - opretter en ny")
    defaultConfig = configparser.ConfigParser()
    with open(script_path + '/kbc-default-config.ini') as defaultCfgFile:
        defaultConfig.read_file(defaultCfgFile)
        config['GENERAL'] = defaultConfig['GENERAL']
        config['EMULATOR'] = defaultConfig['EMULATOR']
        config['MICROBIT'] = defaultConfig['MICROBIT']
        config['LOCALHOST'] = defaultConfig['LOCALHOST']
        config['PRODUCTION'] = defaultConfig['PRODUCTION']

    save()
    print("Afslutter programmet - tjek kbc-config.ini filen og start igen !")
    print('----------------------------------------------------------------')
    exit()


is_production = config['GENERAL'].getboolean('is_production')
is_emulator = config['GENERAL'].getboolean('is_emulator')
poll_interval_in_seconds = config['GENERAL'].getint('polling_interval')
api_user=config['GENERAL']['api_user']
pwd=config['GENERAL']['password']
EMULATOR = config['EMULATOR']
MICROBIT = config['MICROBIT']
LOCALHOST = config['LOCALHOST']
PRODUCTION = config['PRODUCTION']

def get_tty_value(key):
    return EMULATOR[key] if is_emulator else MICROBIT[key]

def get_host_value(key):
    return LOCALHOST[key] if not is_production else PRODUCTION[key]

tty_name = get_tty_value('tty_name')
tty_rate = int(get_tty_value('tty_rate'))
tty_timeout = int(get_tty_value('tty_timeout'))
kbc_host = get_host_value('host')

print(f'  is_production        = {is_production}')
print(f'  is_emulator          = {is_emulator}')
print(f'  tty_name             = {tty_name}')
print(f'  tty_rate             = {tty_rate}')
print(f'  tty_timeout          = {tty_timeout}')
print(f'  kbc_host             = {kbc_host}')
print(f'  polling_interval (s) = {poll_interval_in_seconds}')


import sys
import os
import pty
from time import sleep
from threading import Thread

EMULATOR_PROMPT = 'Skriv kommando der skal sendes til controlleren og tryk <Enter>: '

is_ctrl_initialized = False
track_id = 1

def listen_to_controller(master):
    global is_ctrl_initialized, track_id

    print(f'Starter lytte-tråd...')

    if not is_ctrl_initialized:
        print('Venter på controller i baggrunden...')

    while True:
        s_input = os.read(master, 100)
        recieved = str(s_input, 'utf-8').strip()

        if len(s_input) > 0:
            print()
            print('<-- ' + recieved)

            if (recieved == 'init'):
                print('Controlleren er vågnet...! Init-signal modtaget.')
                print(f'Sender bane-ID "{track_id}" til controller')
                text = f'init:{track_id}\n'
                print(f'--> {text}')
                os.write(master, text.encode('utf-8'))
                sleep(0.5)
                s_input = os.read(master, 100)
                if len(s_input) == 0:
                    print("Fejl! Ingen svar modtaget!")
                else:
                    recieved = str(s_input, 'utf-8').strip()
                    print('<-- ' + recieved)
                    print('Controller INITIALISERET.')
                    is_ctrl_initialized = True

            elif recieved.startswith('EVENT-TRACKS:'):
                temp = recieved[13:]
                slice = temp.index(':')
                track_id = temp[:slice]
                count = temp[slice + 1:]
                print(f'    Kuglebane: {track_id} - antal beskeder: {count}')
             
            elif recieved.startswith('OK:POP:'):
                temp = recieved[7:]
                slice = temp.index(':')
                from_dep_id = temp[:slice]
                temp = temp[len(from_dep_id)+1:]
                slice = temp.index(':')
                from_track_id = temp[:slice]
                command = temp[slice + 1:]
                print(f'    {command} kommando modtaget fra {from_dep_id} - kuglebane {from_track_id}')
             
            elif (recieved == 'OK:QUIT' or recieved == 'quit'):
                print('QUIT kommando modtaget - stopper lytte-tråden!')
                break

            print()
            print(EMULATOR_PROMPT)

def start(id=1):

    global is_ctrl_initialized, track_id

    track_id = id
    if len(sys.argv) > 1:
        track_id = sys.argv[1]

    master, ubitSlave = pty.openpty()
    s_name = os.ttyname(ubitSlave)

    display_emulator_info(s_name)

    event_listener = Thread(group=None, target=listen_to_controller, args=(master,))
    event_listener.start()
    print()

    print()
    print(EMULATOR_PROMPT)
    while True:
        text = input()

        if text == 'clear':
            os.system('cls' if os.name == 'nt' else 'clear')
            display_emulator_info(s_name)
        else:
            if not is_ctrl_initialized:
                print()
                print('Vi venter stadig på controlleren...')
                print()
            else:
                text2send = text + '\n'
                print('--> ' + text)
                os.write(master, text2send.encode('utf-8'))
                sleep(1)

            if  text == 'quit':
                print('Emulatoren lukker og slukker... FARVEL!')
                os.write(master, text.encode('utf-8'))
                break
            

def display_emulator_info(s_name):
    print()
    print()
    print('=============================')
    print('     micro:bit emulator')
    print('=============================')
    print()
    print('Skriv kommandoen "quit" for at afslutte')
    print()
    print('Brug denne port på controlleren:')
    print('--------------------------------')
    print()
    print(s_name)
    print()
    print('--------------------------------')
    print()
    print(f'micro:bit track-ID: {track_id}')
    print()

if __name__ == '__main__':
    start()



from time import sleep
import serial
from . import kb_config

ser = None

def send_message(msg):
    global ser
    if ser is None:
        initialize_connection()

    print('m:b >>> ' + msg)
    ser.write(bytes(msg, 'utf-8'))
    # ensure that the controller do not pick up the message itself 
    sleep(0.2)

def read_line():
    if ser is None:
        initialize_connection()

    text = ser.readline() 
    ser.reset_input_buffer()
    return text

def initialize_connection():
    global ser
    try:
        ser = serial.Serial(kb_config.tty_name, kb_config.tty_rate, timeout=kb_config.tty_timeout)
        ser.reset_input_buffer()
        ser.reset_output_buffer()
        print()
        print(f'Forbinder til micro:bit på: {kb_config.tty_name}')
        sleep(0.2)
        send_message('init')
        return ser
    except Exception as e:
        print()
        print(
            f"Der skete en fejl under forbindelse til seriel porten: {kb_config.tty_name}")
        print("Tjek om microbit'en er forbundet, eller om portnavnet er korrekt.")
        print()
        print(type(e))
        print('----')
        print(e.args)
        exit()

def close_connection():
    if not ser is None:
        ser.close()



from threading import Thread
from time import sleep
from datetime import datetime
from . import microbit
from . import kb_config
from . import kb_central

stop_thread = False
thread = None

def kb_central_poller(current_track_id):
    sleep(1) # Give controller time to initialize
    global stop_thread
    while True:
        print()
        print('- - - - - - - - - - - - - - - - - - - - -')
        print(f"{datetime.now()} POLLER CENTRAL...")

        count = kb_central.get_number_of_events_for_track(current_track_id)
        microbit.send_message(f'EVENT-TRACKS:{current_track_id}:{count}')

        print(f'Kontakter kuglebane-centralen igen om {kb_config.poll_interval_in_seconds} sekunder...')
        print('- - - - - - - - - - - - - - - - - - - - -')
        print()
        sleep(kb_config.poll_interval_in_seconds)
        if (stop_thread):
            print('QUIT signal modtaget - stopper med at polle kuglebane controlleren!')
            break

def start_thread(current_track_id):
    print()
    print(f'Starter tråd der poller kb-centralen: {kb_config.kbc_host}')
    global thread
    if thread is None and not stop_thread:
        thread = Thread(group=None, target=kb_central_poller, args=(current_track_id,))
        thread.start()

def signal_stop_thread():
    global stop_thread
    stop_thread = True


import jwt
from requests.auth import HTTPBasicAuth
import requests
from datetime import datetime
from . import kb_config

jwt_token = None

def get_tracks():
    response = get_request_to_host('api/tracks/all')
    return response.json()['departments']


def send_start_event(from_track_id, to_track_id):
    #'/event/<int:from_track_id>/<int:to_track_id>/<command_id>'
    response = post_request_to_host(f'api/event/{from_track_id}/{to_track_id}/START')
    print(response.text)


def get_events_for_department():
    response = get_request_to_host('api/events/department')
    events = response.json()
    print('Response:')
    print(events)
    return events


def get_number_of_events_for_track(track_id):
    response = get_events_for_track(track_id)
    count = 0 if len(response)==0 else len(response["events"])
    return count

def get_events_for_track(track_id):
    response = get_request_to_host(f'api/events/{track_id}')
    events = response.json()
    print('Response:')
    print(events)
    return events

def delete_event(event_id):
    response = delete_request_to_host(f'api/event/{event_id}')
    print(response.text)


def get_jwt_bearer_auth_header():
    headers = {"Authorization": f"Bearer {get_auth_jwt_token()}"}
    return headers

def get_department_id_from_token():
    return get_decoded_jwt(get_auth_jwt_token())['department_id']

def get_decoded_jwt(token):
    key = "jwt_lkasjdf/(asdkjfh#AQWD!)" if kb_config.is_production else "jwt_dev"
    decoded_jwt = jwt.decode(token, key, algorithms=["HS256"])
    return decoded_jwt

def get_auth_jwt_token():
    global jwt_token

    if jwt_token is None:
        print('Getting token...')
        jwt_token = request_token_from_kb_cetral()

    try:
        get_decoded_jwt(jwt_token)
    except jwt.ExpiredSignatureError:
        print('Token expired! Renewing...')
        jwt_token = request_token_from_kb_cetral()

    return jwt_token

def request_token_from_kb_cetral():
    basic = HTTPBasicAuth(kb_config.api_user, kb_config.pwd)
    url = f'{kb_config.kbc_host}auth/token'
    print(f'KB-CENTRAL >>> {url}')
    try:
        response = requests.get(url, auth=basic)
    except requests.exceptions.ConnectionError as e:
        print_http_connection_error(e)
        exit()
    validate_response(response)
    return response.json()['token']


def validate_response(response):
    if not response.status_code == 200:
        raise Exception(f'Response not OK! URL: {response.url} (status = {response.status_code})')
    
    print(f'KB-CENTRAL <<< Response status OK!')
    print()

    
def get_request_to_host(endpoint):
    url = f'{kb_config.kbc_host}{endpoint}'
    print(f'KB-CENTRAL >>> {url}  [GET]')
    try:
        response = requests.get(url, headers=get_jwt_bearer_auth_header())
    except requests.exceptions.ConnectionError as e:
        print_http_connection_error(e)
        exit()
    validate_response(response)
    return response
    
def post_request_to_host(endpoint):
    url = f'{kb_config.kbc_host}{endpoint}'
    print(f'KB-CENTRAL >>> {url}  [POST]')
    try:
        response = requests.post(url, headers=get_jwt_bearer_auth_header())
    except requests.exceptions.ConnectionError as e:
        print_http_connection_error(e)
        exit()
    validate_response(response)
    return response

def delete_request_to_host(endpoint):
    url = f'{kb_config.kbc_host}{endpoint}'
    print(f'KB-CENTRAL >>> {url}  [DELETE]')
    try:
        response = requests.delete(url, headers=get_jwt_bearer_auth_header())
    except requests.exceptions.ConnectionError as e:
        print_http_connection_error(e)
        exit()
    validate_response(response)
    return response

def print_http_connection_error(e):    
    print()
    print("FEJL! Ingen forbindelse til kugelbane centralen!")
    print(f"Host: {kb_config.kbc_host}")
    print()
    print(type(e))
    print(e)
    print(e.args)
